"""Definition for main data models used in this library."""

from dataclasses import dataclass
from enum import IntEnum, auto


class EdgeQLOperationType(IntEnum):
    """Enumeration for operation types for queries."""

    #: type for operation that returns a single object.
    single_return = auto()

    #: type for operation that returns a common set of object.
    set_return = auto()

    #: type for operation that returns nothing.
    execute = auto()


@dataclass
class Query:
    """Parsed query."""

    #: name of parsed query.
    name: str

    #: query operation type.
    operation_type: EdgeQLOperationType

    #: EdgeQL query that should be executed.
    edgeql: str

    def __hash__(self) -> int:
        """Hash query.

        Hashing is done by query's name.

        Returns:
            Query's hash.
        """
        return hash(self.name)

    def __str__(self) -> str:
        """Return string representation of query.

        Returns:
            String representation of query that is its name.
        """
        return self.name

    def __repr__(self) -> str:
        """Return raw string representation of query.

        Returns:
            Raw string representation of query that contains all fields.
        """
        return "Query(name: {0!r}, operation_type: {1}, edgeql: {2!r})".format(
            self.name,
            self.operation_type.name,
            self.edgeql,
        )



"""Definition for aiosql compatible adapters.

Adapters here aren't "real" adapters, that can execute queries
but just dummy classes with compatible interface.
"""


# this drivers won't and shouldn't be used as real drivers.
# they are required only for queries building.


class EdgeQLAsyncAdapter:
    """Dummy adapter for async driver."""

    is_aio_driver = True


class EdgeQLSyncAdapter:
    """Dummy adapter for sync driver."""

    is_aio_driver = False



"""Definition for main collection for queries."""

from __future__ import annotations

from typing import Callable, Dict, List, Set, Union

from edgeql_queries.executors.async_executor import create_async_executor
from edgeql_queries.executors.sync_executor import create_sync_executor
from edgeql_queries.models import Query
from edgeql_queries.typing import QueriesTree


def _create_handler_from_query(query: Query, use_async: bool = True) -> Callable:
    if use_async:
        return create_async_executor(query)

    return create_sync_executor(query)


def load_from_list(queries_collection: Queries, queries: List[Query]) -> Queries:
    """Add queries from list.

    Arguments:
        queries_collection: already registered queries.
        queries: list of queries to be added.

    Returns:
        Collection of queries to which method was applied.
    """
    for query in queries:
        queries_collection.add_query(query.name, query)

    return queries_collection


def load_from_tree(queries_collection: Queries, query_tree: QueriesTree) -> Queries:
    """Add queries from tree.

    Arguments:
        queries_collection: already registered queries.
        query_tree: tree of queries that should be added.

    Returns:
        Collection of queries to which method was applied.
    """
    for group_name, queries in query_tree.items():
        if isinstance(queries, dict):
            queries_collection.add_query(
                group_name,
                load_from_tree(Queries(queries_collection.is_async), queries),
            )
        else:
            queries_collection.add_query(queries.name, queries)

    return queries_collection


class Queries:
    """Collection and executor for queries."""

    def __init__(self, is_async: bool = True) -> None:
        """Initialize collection and executor for queries.

        Arguments:
            is_async: use async driver for creating queries.
        """
        self._query_handlers: Dict[str, Union[Callable, "Queries"]] = {}
        self._available_queries: Set[Query] = set()
        self._available_queries_groups: Dict[str, Queries] = {}
        self._is_async = is_async

    @property
    def available_queries(self) -> List[Query]:
        """Sorted list of queries available on this collection.

        Returns:
            List of queries.
        """
        return sorted(self._available_queries, key=lambda query: query.name)

    @property
    def is_async(self) -> bool:
        """Will be query handlers generated for async execution.

        Returns:
            Will be query handlers generated for async execution.
        """
        return self._is_async

    def add_query(self, name: str, query_handler: Union[Queries, Query]) -> None:
        """Add a single query to collection.

        Arguments:
            name: name of query or sub-queries to be added.
            query_handler: a single [query][edgeql_queries.models.Query] that
                will be transformed to executor or
                [collection of queries][edgeql_queries.queries.Queries]
                that will be registered as sub-queries.
        """
        handler_for_query: Union[Callable, Queries]

        if isinstance(query_handler, Query):
            self._available_queries.add(query_handler)

            handler_for_query = _create_handler_from_query(
                query_handler,
                self._is_async,
            )
        else:
            handler_for_query = query_handler
            self._available_queries_groups[name] = handler_for_query

        self._query_handlers[name] = handler_for_query

    def get_executor(self, query_name: str) -> Union[Callable, "Queries"]:
        """Return executor for query by name.

        Arguments:
            query_name: name of query for which executor should be returned.

        Returns:
            Executor for query.
        """
        return self._query_handlers[query_name]

    def __getattr__(self, query_name: str) -> Union[Callable, "Queries"]:
        """Get executor for query by name.

        Arguments:
            query_name: name of query or group.

        Returns:
            Executor for query.
        """
        return self.get_executor(query_name)

    def __repr__(self) -> str:
        """Return special string representation of collection.

        Returns:
            Raw string for queries collection.
        """
        return "Queries(queries: {0}, groups: {1})".format(
            self.available_queries,
            self._available_queries_groups,
        )



"""Definition for parsing functions for queries.

Queries that are acceptable by this library should follow some rules:

1. They should be named.
2. Their names should not contain symbols that could not
    be used in Python identifier(**except for `-` symbol since
    it would be converted into `_`**)
3. They can have special symbols after their names that will change how
    this queries will be executed:
    * `*`: query will be executed as script with using `.execute` method from driver.
    * `!`: query will always return a single object and executed with `.fetchone`
        method from driver.
    * empty: common query that will return a set of objects and will be executed with
        `.fetchall` method from driver.

An example query that can be successfuly parsed:
```edgeql
    # name: select-user-by-username!
    # Find user by username and return it.
    SELECT User {
        username,
        bio,
        is_active
    }
    FILTER .username = <str>$username
    LIMIT 1
```
"""

import re
from types import MappingProxyType
from typing import Mapping, Tuple

from edgeql_queries.exceptions import EdgeQLParsingError
from edgeql_queries.models import EdgeQLOperationType, Query

QUERY_DEFINITION_PATTERN = re.compile(r"^\s*#\s*name:\s*(.*)\s*", re.MULTILINE)
VALID_QUERY_NAME_PATTERN = re.compile(r"^\w+$")

_OPERATION_SUFFFIXES_TO_TYPES: Mapping[str, EdgeQLOperationType] = MappingProxyType(
    {
        "*": EdgeQLOperationType.execute,
        "!": EdgeQLOperationType.single_return,
        "": EdgeQLOperationType.set_return,
    },
)


def get_query_name_and_operation(name: str) -> Tuple[str, EdgeQLOperationType]:
    """Return query name and operation from query headers.

    Arguments:
        name: raw query name with operator from which
            final name and operation should be extracted.

    Returns:
        Query name and [operation type][edgeql_queries.models.EdgeQLOperationType]

    Raises:
        EdgeQLParsingError: if header is in wrong format or name could not be
            converted into Python identificator.
    """
    name = name.replace("-", "_")

    operation_suffix = ""
    for suffix in _OPERATION_SUFFFIXES_TO_TYPES:  # pragma: no branch
        if name.endswith(suffix):
            operation_suffix = suffix
            break

    if operation_suffix:
        query_name = name[: -len(operation_suffix)]
    else:
        query_name = name

    if not VALID_QUERY_NAME_PATTERN.match(query_name):
        raise EdgeQLParsingError(
            'name must be convertable to valid python variable, got "{0}"'.format(
                query_name,
            ),
        )

    return (
        query_name,
        _OPERATION_SUFFFIXES_TO_TYPES[operation_suffix],
    )


def parse_query_from_string(raw_name: str, query_body: str) -> Query:
    """Parse EdgeQL query string into [edgeql_queries.models.Query].

    Arguments:
        raw_name: query name with operation.
        query_body: EdgeQL query.

    Returns:
        [edgeql_queries.models.Query] that will be later added for creating executors.
    """
    query_name, operation_type = get_query_name_and_operation(raw_name)

    return Query(query_name, operation_type, query_body)



"""Definition of classes for usage with aiosql."""

from edgeql_queries.contrib.aiosql.adapters import EdgeQLAsyncAdapter, EdgeQLSyncAdapter
from edgeql_queries.contrib.aiosql.queries import EdgeQLQueries
from edgeql_queries.contrib.aiosql.query_loaders import EdgeQLQueryLoader

__all__ = (
    "EdgeQLAsyncAdapter",
    "EdgeQLSyncAdapter",
    "EdgeQLQueries",
    "EdgeQLQueryLoader",
)



"""Definition of loader for queries from files and directories."""

from pathlib import Path
from typing import Iterator, List, Match, Optional, Tuple

from edgeql_queries.models import Query
from edgeql_queries.parsing import QUERY_DEFINITION_PATTERN, parse_query_from_string
from edgeql_queries.typing import QueriesTree

MatchPair = Tuple[Match, Optional[Match]]


def _iter_pairs(matches: Iterator[Match]) -> Iterator[MatchPair]:
    start_match = next(matches)

    try:
        end_match = next(matches)
    except StopIteration:
        yield (start_match, None)
        return

    yield (start_match, end_match)

    for next_match in matches:
        start_match, end_match = end_match, next_match
        yield (start_match, end_match)

    yield (end_match, None)


def load_query_data_from_edgeql(edgeql: str) -> List[Query]:
    """Load queries from string.

    Arguments:
        edgeql: edgeql string that contains queries to be parsed.

    Returns:
        List of [queries][edgeql_queries.models.Query] that will be used
            later by [container][edgeql_queries.queries.Queries] for them.
    """
    query_data = []
    matches_iter = _iter_pairs(QUERY_DEFINITION_PATTERN.finditer(edgeql))
    for (start_match, end_match) in matches_iter:
        if end_match is not None:
            end_position = end_match.start()
        else:
            end_position = len(edgeql)
        query_data.append(
            parse_query_from_string(
                start_match.groups()[0],
                edgeql[start_match.end() : end_position],
            ),
        )
    return query_data


def load_query_data_from_file(file_path: Path) -> List[Query]:
    """Load queries from esdl file.

    Arguments:
        file_path: path to file with queries to be parsed.

    Returns:
        List of [queries][edgeql_queries.models.Query] that will be used
            later by [container][edgeql_queries.queries.Queries] for them.

    Raises:
        FileNotFoundError: if failed to open file.  # noqa: DAR402
    """
    with open(file_path) as edgeql_file:
        return load_query_data_from_edgeql(edgeql_file.read())


def load_query_data_from_dir_path(dir_path: Path) -> QueriesTree:
    """Load queries from esdl file.

    Arguments:
        dir_path: path to file with queriesto be parsed.

    Returns:
        List of [queries][edgeql_queries.models.Query] that will be used
            later by [container][edgeql_queries.queries.Queries] for them.

    Raises:
        ValueError: if dir_path is not directory.
    """
    if not dir_path.is_dir():
        raise ValueError("path {0} must be a directory".format(dir_path))

    return _load_query_data_tree(dir_path, dir_path)


def _load_query_data_tree(root_path: Path, path: Path) -> QueriesTree:
    query_data_tree: QueriesTree = {}
    for sub_path in path.iterdir():
        if sub_path.is_file() and sub_path.suffix == ".edgeql":
            for query in load_query_data_from_file(sub_path):
                query_data_tree[query.name] = query
        elif sub_path.is_dir():
            group_name = sub_path.relative_to(root_path).name
            query_data_tree[group_name] = _load_query_data_tree(root_path, sub_path)
    return query_data_tree



"""Definition for creator of sync executors."""

from functools import partial
from types import MappingProxyType
from typing import Any, Callable, Mapping

from edgedb import BlockingIOConnection
from edgedb.datatypes import datatypes

from edgeql_queries.models import EdgeQLOperationType, Query


def _execute(__edgeql_query__: Query, conn: BlockingIOConnection) -> None:
    return conn.execute(__edgeql_query__.edgeql)


def _set_return(
    __edgeql_query__: Query,
    conn: BlockingIOConnection,
    *query_args: Any,
    **query_kwargs: Any,
) -> datatypes.Set:
    return conn.query(__edgeql_query__.edgeql, *query_args, **query_kwargs)


def _single_return(
    __edgeql_query__: Query,
    conn: BlockingIOConnection,
    *query_args: Any,
    **query_kwargs: Any,
) -> Any:
    return conn.query_one(__edgeql_query__.edgeql, *query_args, **query_kwargs)


_OPERATION_TO_EXECUTOR: Mapping[EdgeQLOperationType, Callable] = MappingProxyType(
    {
        EdgeQLOperationType.single_return: _single_return,
        EdgeQLOperationType.set_return: _set_return,
        EdgeQLOperationType.execute: _execute,
    },
)


def create_sync_executor(query: Query) -> Callable:
    """Create sync executor for query.

    Arguments:
        query: query for which executor should be created.

    Returns:
        Created sync executor.
    """
    executor = _OPERATION_TO_EXECUTOR[query.operation_type]
    return partial(executor, query)



"""Definition of loaders for queries."""

from pathlib import Path
from typing import Union

from edgeql_queries.exceptions import EdgeQLLoadError
from edgeql_queries.queries import Queries, load_from_list, load_from_tree
from edgeql_queries.query_loaders import (
    load_query_data_from_dir_path,
    load_query_data_from_file,
)


def from_path(edgeql_path: Union[str, Path], async_driver: bool = True) -> Queries:
    """Load queries by path to `.edgeql` file or directory with them.

    Arguments:
        edgeql_path: path to `.edgeql` file with queries or directory
            with queries.
        async_driver: create async executors for using with async driver.

    Returns:
        Loaded and parsed queries.

    Raises:
        EdgeQLLoadError: if path to queries does not exist.
    """
    path = Path(edgeql_path)

    if not path.exists():
        raise EdgeQLLoadError("{0} does not exist".format(path))

    queries = Queries(async_driver)

    if path.is_file():
        query_data = load_query_data_from_file(path)
        return load_from_list(queries, query_data)
    elif path.is_dir():
        query_data_tree = load_query_data_from_dir_path(path)
        return load_from_tree(queries, query_data_tree)

    raise EdgeQLLoadError(
        "edgeql_path must be a directory or file, got {0}".format(edgeql_path),
    )



"""Definition for aiosql compatible queries."""

from typing import List, Union

from edgeql_queries import queries as eq_queries
from edgeql_queries.contrib.aiosql.adapters import EdgeQLAsyncAdapter, EdgeQLSyncAdapter
from edgeql_queries.models import Query
from edgeql_queries.typing import QueriesTree


class EdgeQLQueries:
    """Queries that are compatible with aiosql."""

    def __init__(self, adapter: Union[EdgeQLSyncAdapter, EdgeQLAsyncAdapter]) -> None:
        """Init queries.

        Arguments:
            adapter: adapter for aiosql with `is_aio_driver` field.
        """
        self._use_async = adapter.is_aio_driver

    def load_from_list(self, queries: List[Query]) -> eq_queries.Queries:
        """Load list of queries.

        Arguments:
            queries: list of queries that should be used for creating
                executors for them.

        Returns:
            Built collection of queries with binded executors.
        """
        return eq_queries.load_from_list(eq_queries.Queries(self._use_async), queries)

    def load_from_tree(self, queries_tree: QueriesTree) -> eq_queries.Queries:
        """Load queries tree.

        Arguments:
            queries_tree: tree of queries that should be used for creating
                executors for them.

        Returns:
            Built collection of queries with binded executors.
        """
        return eq_queries.load_from_tree(
            eq_queries.Queries(self._use_async),
            queries_tree,
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['edgeql_queries',
 'edgeql_queries.contrib',
 'edgeql_queries.contrib.aiosql',
 'edgeql_queries.executors']

package_data = \
{'': ['*']}

install_requires = \
['edgedb>=0.9.0']

setup_kwargs = {
    'name': 'edgeql-queries',
    'version': '0.0.5',
    'description': 'Simple EdgeQL in Python.',
    'long_description': '<h1 align="center">edgeql-queries</h1>\n<p align="center">\n    <em>Simple EdgeQL in Python.</em>\n</p>\n<p align="center">\n    <a href=https://github.com/nsidnev/edgeql-queries>\n        <img src=https://github.com/nsidnev/edgeql-queries/workflows/Tests/badge.svg alt="Tests" />\n    </a>\n    <a href=https://github.com/nsidnev/edgeql-queries>\n        <img src=https://github.com/nsidnev/edgeql-queries/workflows/Styles/badge.svg alt="Styles" />\n    </a>\n    <a href="https://codecov.io/gh/nsidnev/edgeql-queries">\n        <img src="https://codecov.io/gh/nsidnev/edgeql-queries/branch/master/graph/badge.svg" alt="Coverage" />\n    </a>\n    <a href="https://github.com/ambv/black">\n        <img src="https://img.shields.io/badge/code%20style-black-000000.svg" alt="Code Style" />\n    </a>\n    <a href="https://github.com/wemake-services/wemake-python-styleguide">\n        <img src="https://img.shields.io/badge/style-wemake-000000.svg" alt="WPS Linter"/>\n    </a>\n    <a href="https://github.com/nsidnev/edgeql-queries/blob/master/LICENSE">\n        <img src="https://img.shields.io/badge/License-FreeBSD-blue" alt="License" />\n    </a>\n    <a href="https://pypi.org/project/edgeql-queries/">\n        <img src="https://badge.fury.io/py/edgeql-queries.svg" alt="Package version" />\n    </a>\n</p>\n\n---\n\n**Documentation**: https://nsidnev.github.io/edgeql-queries/\n\n## Requirements\n\n`edgeql-queries` requires only the [EdgeDB driver for Python](https://github.com/edgedb/edgedb-python).\n\n## Credits\n\nThis project is inspired by [aiosql](https://github.com/nackjicholson/aiosql)\nproject and is based on it\'s source code.\n\n## License\n\nThis project is licensed under the terms of the FreeBSD license.\n',
    'author': 'Nik Sidnev',
    'author_email': 'sidnev.nick@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/nsidnev/edgeql-queries',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Definition for aiosql compatible query loader."""

from pathlib import Path
from typing import Any, List

from edgeql_queries import query_loaders
from edgeql_queries.models import Query
from edgeql_queries.typing import QueriesTree


class EdgeQLQueryLoader:
    """Query loader that is compatible with aiosql."""

    def __init__(self, *args: Any) -> None:
        """Do nothing but be compatible with aiosql query loader.

        Arguments:
            args: any args that will be used for initializaton by aiosql.
        """

    def load_query_data_from_sql(self, edgeql: str) -> List[Query]:
        """Load queries from EdgeQL string.

        Arguments:
            edgeql: EdgeQL string to create a new collection of queries.

        Returns:
            Parsed queries.
        """
        return query_loaders.load_query_data_from_edgeql(edgeql)

    def load_query_data_from_file(self, file_path: Path) -> List[Query]:
        """Load queries from single file.

        Arguments:
            file_path: path to EdgeQL file to create a new collection of
                queries.

        Returns:
            Parsed queries.
        """
        return query_loaders.load_query_data_from_file(file_path)

    def load_query_data_from_dir_path(self, dir_path: Path) -> QueriesTree:
        """Load queries from directory.

        Arguments:
            dir_path: path to dir with EdgeQL files to create a new collection
                of queries.

        Returns:
            Parsed queries.
        """
        return query_loaders.load_query_data_from_dir_path(dir_path)



"""Definition for aliases for complex types from typing."""

from typing import Dict, Union

from edgeql_queries.models import Query

# mypy has limitation on understanding cyclic types
# https://github.com/python/mypy/issues/7069
QueriesTree = Dict[str, Union[Query, "QueriesTree"]]  # type: ignore



"""Definition for error that may be raised by this library."""


class EdgeQLLoadError(Exception):
    """Error that raised if there is an error to load queries."""


class EdgeQLParsingError(Exception):
    """Error that raised if there is an error in parsing query."""



"""Definition for creator of async executors."""

from functools import partial
from types import MappingProxyType
from typing import Any, Callable, Mapping, Union

from edgedb import AsyncIOConnection, AsyncIOPool
from edgedb.datatypes import datatypes

from edgeql_queries.models import EdgeQLOperationType, Query

_AsyncFetcher = Union[AsyncIOConnection, AsyncIOPool]


async def _execute(__edgeql_query__: Query, conn: _AsyncFetcher) -> None:
    return await conn.execute(__edgeql_query__.edgeql)


async def _set_return(
    __edgeql_query__: Query,
    conn: _AsyncFetcher,
    *query_args: Any,
    **query_kwargs: Any,
) -> datatypes.Set:
    return await conn.query(__edgeql_query__.edgeql, *query_args, **query_kwargs)


async def _single_return(
    __edgeql_query__: Query,
    conn: _AsyncFetcher,
    *query_args: Any,
    **query_kwargs: Any,
) -> Any:
    return await conn.query_one(__edgeql_query__.edgeql, *query_args, **query_kwargs)


_OPERATION_TO_EXECUTOR: Mapping[EdgeQLOperationType, Callable] = MappingProxyType(
    {
        EdgeQLOperationType.single_return: _single_return,
        EdgeQLOperationType.set_return: _set_return,
        EdgeQLOperationType.execute: _execute,
    },
)


def create_async_executor(query: Query) -> Callable:
    """Create async executor for query.

    Arguments:
        query: query for which executor should be created.

    Returns:
        Created async executor.
    """
    executor = _OPERATION_TO_EXECUTOR[query.operation_type]
    return partial(executor, query)
